#include "common.h"

void err_exit(char *s) {
	printf("%s\n",s);
	exit(0);
}

void showversion() {
	printf("\nhwing version %s, by snooq [jinyean@hotmail.com]\n",VERSION);
	exit(0);
}

void usage(char *s) {
	printf("\nUsage: %s [options] host\n\n",s);
	printf("\t-h\tShow this help menu\n");
	printf("\t-v\tShow version number\n");
	printf("\t-c\tStop after count response packets.\n");
	printf("\t-i\tInterval in X (seconds) or uX (micro seconds)\n");
	printf("\t-M\tSet sequence number\n");
	printf("\t-a\tSpoof source address\n");
	printf("\t-p\tDestination port\n");
	printf("\t-s\tSource port\n");
	printf("\t-R\tSet RST tcp flag\n");
	printf("\t-S\tSet SYN tcp flag\n");
	printf("\t-A\tSet ACK tcp flag\n");
	printf("\t-F\tSet FIN tcp flag\n");
	printf("\t-P\tSet PSH tcp flag\n");
	printf("\t-U\tSet URG tcp flag\n");
	exit(0);
}

void cleanup() {
	closesocket(rd);
	closesocket(sd);
	WSACleanup();
	exit(0);
}

void intrpt() {
	float lc=0;
	if (rc>sc) { err_exit("We get more than what we sent??!??"); }
	if (rc==0) {
		lc=100;
		min_rtt=0;
	} 
	else {
		lc=(float)(100-(((float)rc/(float)sc)*100));
		avg_rtt=tmp_rtt/(float)rc;
	}

    printf("\n---- %s hwing statistic ----\n",host);
	printf("%d packets transmitted, %d packets received, %.1f%% packet loss\n",sc,rc,lc);
	printf("round-trip min/avg/max = %.2f/%.2f/%.2f ms\n",min_rtt,avg_rtt,max_rtt);
	cleanup();
}

// Standard TCP/UDP checksum function.
__u16 checksum(__u16 *buffer, int size) {
    unsigned long cksum=0;
    while (size > 1) {
        cksum += *buffer++;
        size  -= sizeof(__u16);   
    }
    if (size) {
        cksum += *(__u16*)buffer;   
    }
    cksum = (cksum >> 16) + (cksum & 0xffff);
    cksum += (cksum >>16); 
    return (__u16)(~cksum); 
}


void fill_ip_hdr (char *packet,__u8 protocol,__u32 sa,__u32 da,__u16 len) {

	IP_HDR *ipHdr;
	__u16 IP_Size;

	ipHdr=(IP_HDR*)packet;
	memset((char *)ipHdr,'\0',sizeof(IP_HDR));
  
   	IP_Size=sizeof(IP_HDR)/sizeof(__u32);

	// Setup our IP header
   	ipHdr->ip_verlen = (IP_VERSION << 4) | IP_Size;
   	ipHdr->ip_tos = 0;					// IP type of service
   	ipHdr->ip_totallength = len;		// Total packet len
   	ipHdr->ip_id = 1;					// Unique identifier: set to 0
	ipHdr->ip_offset = 0;				// Fragment offset field
   	ipHdr->ip_ttl = DEFAULT_TTL;		// Time to live
   	ipHdr->ip_protocol = protocol;		// Protocol(TCP/UDP) 
   	ipHdr->ip_checksum = checksum((__u16 *)ipHdr,sizeof(IP_HDR));    // IP checksum
	ipHdr->ip_srcaddr = sa;				// Source address
   	ipHdr->ip_destaddr = da;			// Destination address

}


/* Compute TCP checksum with pseudo header. */

__u16 tcp_checksum(__u8 proto,char *packet,__u16 len,__u32 sa,__u32 da) {

	HANDLE hHeap;
	PSD_HDR pseudo_hdr;
	char *pseudo_pkt;
	__u16 answer;
  
	hHeap=GetProcessHeap();

	pseudo_hdr.protocol = proto;
	pseudo_hdr.tcp_length = htons(len);
	pseudo_hdr.placeholder = 0;
	pseudo_hdr.source_address = sa;
	pseudo_hdr.dest_address = da;
  
	if((pseudo_pkt=HeapAlloc(hHeap,0,sizeof(PSD_HDR)+len))==NULL) {
		_snprintf(tmp,GP_BSIZE-1,"HeapAlloc error!\n");
		err_exit(tmp);
	}
  
	memcpy(pseudo_pkt,&pseudo_hdr,sizeof(pseudo_hdr));
	memcpy((pseudo_pkt+sizeof(pseudo_hdr)),packet,len);
  
	answer=(__u16)checksum((__u16 *)pseudo_pkt,(len+sizeof(pseudo_hdr)));

	HeapFree(hHeap,0,pseudo_pkt);
	
	return answer;
}

void fill_tcp_hdr(char *packet,__u16 s_port,__u16 d_port,__u32 sa,
				  __u32 da, __u32 c_seq, __u32 s_seq,__u8 flags) {

 	TCP_HDR *tcp;

	tcp=(TCP_HDR *)packet;
	memset((char *)tcp,'\0',sizeof(TCP_HDR));

	// Fill up our TCP header
    tcp->sport = htons(s_port);			
    tcp->dport = htons(d_port) ;		
	tcp->seqnum = htonl(c_seq);	
	tcp->acknum = htonl(s_seq);
	tcp->DataOffset=(TH_OFFSET) << 4;

	// Set SYN flag on.
	tcp->Flags=flags;
	tcp->Windows=htons(TCP_WINDOW_SIZE);
	tcp->Checksum=tcp_checksum(IPPROTO_TCP,(char *)tcp,sizeof(TCP_HDR),sa,da);
	tcp->UrgPointer=0;

}

void hex_dump(__u8 *packet, int size) {
	unsigned char *byte = packet;
	int count = 0;

	printf("\t\t");
	for (; byte < (unsigned char*) (packet+size); byte++) {
		count++;
		printf("%02x", *byte);
		if (count % 2 == 0) printf(" ");
		if (count % 16 == 0) printf("\n\t\t");
	}
	printf("\n\n");
}

void human_dump(__u8 *packet, int size) {
	unsigned char *byte = packet;
	int count = 0;

	printf("\t\t");
	for (; byte < (unsigned char*) (packet+size); byte++) {
		count ++;
		if (isprint(*byte))
			printf("%c", *byte);
		else
			printf(".");
		if (count % 32 == 0) printf("\n\t\t");
	}
	printf("\n\n");
}

void get_bytes(long addr) {

	b[0]=addr&0xff;
    b[1]=(addr>>8)&0xff;
    b[2]=(addr>>16)&0xff;
    b[3]=(addr>>24)&0xff;

}

void init_socks() {

	BOOL on=TRUE;

	if (QueryPerformanceFrequency((LARGE_INTEGER *)&freq)==0) err_exit("Can't initialise timer.");

	if (WSAStartup(MAKEWORD(2,2),&wsd)!=0) {
        _snprintf(tmp,GP_BSIZE-1,"WSAStartup() failed: %d", GetLastError());
		err_exit(tmp);
	}

	sd=WSASocket(AF_INET, SOCK_RAW, IPPROTO_RAW, NULL, 0,0);

	if (sd==INVALID_SOCKET) {
        _snprintf(tmp,GP_BSIZE-1,"WSASocket() failed: %d", WSAGetLastError());
		err_exit(tmp);
    }
  
	if (setsockopt(sd,IPPROTO_IP,IP_HDRINCL,(char *)&on,sizeof(on)) < 0) {
		_snprintf(tmp,GP_BSIZE-1,"setsockopt(IP_HDRINCL) failed: %d\n", WSAGetLastError());
		err_exit(tmp);
	}

	rd=WSASocket(AF_INET, SOCK_RAW, IPPROTO_IP, NULL, 0,0);

	if (rd==INVALID_SOCKET) {
        _snprintf(tmp,GP_BSIZE-1,"WSASocket() failed: %d", WSAGetLastError());
		err_exit(tmp);
    }
  
}

char *getmyip(int e) {
	int i;
	char ac[MAX_HOSTNAME_LAN];
	struct in_addr addr;
	struct hostent *phe;

	if (gethostname(ac, sizeof(ac))==SOCKET_ERROR) {
		err_exit("Couldn't get local hostname...");
	}
  
	phe=gethostbyname(ac);

	if (phe==0) {
        	err_exit("Hostname lookup error...");
	}

	if (e) {
		for (i=0; phe->h_addr_list[i]!=0; ++i) {
			memcpy(&addr, phe->h_addr_list[i], sizeof(struct in_addr));
			printf("-> Use %s as local IP? [y/n]\n",inet_ntoa(addr));
			if (getch()=='y' || getch()=='Y') return inet_ntoa(addr);
		}
		err_exit("Couldn't find local IP??? Try '-i' switch...\n");
	}
	else {
		memcpy(&addr, phe->h_addr_list[0], sizeof(struct in_addr));
		return inet_ntoa(addr);
	}

	return NULL;
}

void send_packet(__u32 sa,__u32 sp, __u32 da,__u32 dp,__u32 cs, __u32 ss,__u8 flags) {

	__u8 packet[sizeof(IP_HDR)+sizeof(TCP_HDR)];

	SOCKIN sock_addr;
	TCP_HDR *tcp;
	int retn=0,i=0;

	fill_ip_hdr(packet,IPPROTO_TCP,sa,da,sizeof(packet));
 
	tcp=(TCP_HDR *)(packet+sizeof(IP_HDR));

	fill_tcp_hdr((char *)tcp,sp+i,dp,sa,da,cs,ss,flags);
  
	memset(&sock_addr,'\0',sizeof(sock_addr));
  
	sock_addr.sin_family=AF_INET;
	sock_addr.sin_port=htons(dp);
	sock_addr.sin_addr.s_addr=da;

	retn=sendto(sd,packet,sizeof(packet),0,(SOCKADDR *)&sock_addr,sizeof(sock_addr));
	
	if (QueryPerformanceCounter((LARGE_INTEGER *)&start)==0) err_exit("Counter failed");

	if (retn==SOCKET_ERROR) {
		_snprintf(tmp,GP_BSIZE-1,"sendto() error: %d",WSAGetLastError());
		err_exit(tmp);
	}

//	printf("sent #%d -> src ip: %x, src port: %d\n",sc,sa,sp);
	sc++;
}

void setpromiscuos(__u32 sa) {

	BOOL on=TRUE;
	SOCKIN sockin;

	char FAR name[MAX_HOSTNAME_LAN];

	gethostname(name,MAX_HOSTNAME_LAN);

	pHostent=(struct hostent * )malloc(sizeof(struct hostent));
	pHostent=gethostbyname(name);

	sockin.sin_family=AF_INET;
	sockin.sin_port=htons(6000);
	memcpy(&sockin.sin_addr.S_un.S_addr, pHostent->h_addr_list[0], pHostent->h_length);

	if(bind(rd,(SOCKADDR *)&sockin,sizeof(sockin))<0) {
		_snprintf(tmp,GP_BSIZE-1,"bind error: %d",WSAGetLastError());
		err_exit(tmp);
	}

	if(ioctlsocket(rd, SIO_RCVALL, &on)<0) {    
		err_exit("Could not set promiscuos mode"); 
	} 

}

char *get_flags(__u8 flags) {
	recvflags[0]='\0';
	if (flags & RST) strcat(recvflags,"R");
	if (flags & SYN) strcat(recvflags,"S");
	if (flags & ACK) strcat(recvflags,"A");
	if (flags & FIN) strcat(recvflags,"F");
	if (flags & PSH) strcat(recvflags,"P");
	if (flags & URG) strcat(recvflags,"U");
	if (recvflags[0]=='\0') strcat(recvflags,"none");
	return recvflags;
}

int wait_packet(__u32 sa, __u16 port, int cnt) {

	IP_HDR *ip;
	TCP_HDR *tcp;

	char buffer[BUFFER_SIZE];
	int size, c=0, received=0;
	double rtt;
		
	__u16 dst_port, src_port;
	__u32 dst_addr, on=1;

	while (!received) {	

		memset(buffer, 0, sizeof(buffer));
		size=recv(rd,buffer,sizeof(buffer),0);

		if (QueryPerformanceCounter((LARGE_INTEGER *)&stop)==0) err_exit("Counter failed");
		elapsed=stop-start;
		rtt=((double)elapsed/(double)freq)*1000;

		if (size==SOCKET_ERROR) {
			_snprintf(tmp,GP_BSIZE-1,"Error recv() = %ld ",WSAGetLastError());
			err_exit(tmp);
		}

	//	printf("recv size: %d\n",size);

		if (size>=(sizeof(IP_HDR)+sizeof(TCP_HDR))) {

			ip=(IP_HDR *)buffer;
			dst_addr=ip->ip_destaddr;
			tcp=(TCP_HDR *)(buffer+sizeof(IP_HDR));   
			dst_port=ntohs(tcp->dport);	
			src_port=ntohs(tcp->sport);	

			//printf("we get -> src ip: %x, src port: %d\n",dst_addr,dst_port);						

			/*
			get_bytes(ip->ip_destaddr);
			printf(" %d.%d.%d.%d:%d\n",b[0],b[1],b[2],b[3],dst_port);
			*/

			if ((sa==dst_addr)&&(port==(__u16)dst_port)) {

				printf("len=%d ",size);
				get_bytes(ip->ip_srcaddr);
				printf("IP=%d.%d.%d.%d",b[0],b[1],b[2],b[3]);
		
				printf(" ttl=%d ",ip->ip_ttl);
				if (ntohs(ip->ip_offset)) printf("DF ");
				printf("id=%d",ip->ip_id);

				printf(" sport=%d",src_port);
				printf(" flags=%s",get_flags(tcp->Flags));
				printf(" seq=%d",rc);
				printf(" rtt=%.2f ms\n",rtt);
				received=1;
				tmp_rtt+=rtt;
				if (rtt>max_rtt) max_rtt=rtt;
				if (rtt<min_rtt) min_rtt=rtt;
				rc++;
				if (rc==cnt) intrpt();
			}
		}
	}

	return 0;
}
